# '{{ project_name }}' on NimbleBox

This is an autogenerated README file for a [NBX](www.nimblebox.ai) app:
* created at: {{ created_time }}
* name: {{ project_name }}
* created by: {{ email_id }}

Steps to complete:

1. Add your code/folders/csv/etc. in this folder and fill requirements.txt with all the dependencies
2. Fill `nbx_user.py:get_operator` function and `python3 exe.py deploy`
3. Sit back, relax, and strap on your seat belt. - Dr. Dre

Protip: Create a new file with code for your job and import that inside the `nbx_user.get_op()` function
