# This is an autogenerated nbx_user.py file for a [NBX](www.nimblebox.ai) app:
# project name: {{ project_name }}
# created time: {{ created_time }}
#   created by: {{ email_id }}

import os

#nbx nbox_setup/

os.environ["NBOX_LOG_LEVEL"] = "INFO" # Keep it the way you like

from nbox import Operator
from nbox.hyperloop.job_pb2 import Resource

def get_op() -> Operator:

  job = None # <<< YOUR Operator HERE #

  # confused? uncomment lines below
  # from nbox.nbxlib.ops import Magic
  # job = Magic()

  return job

def get_resource() -> Resource:
  """Define your pod config here"""
  return Resource(
    cpu = "100m",         # 100mCPU
    memory = "200Mi",     # MiB
    disk_size = "1Gi",    # GiB
    gpu = "none",         # keep "none" for no GPU
    gpu_count = "0",      # keep "0" when no GPU
    timeout = 300_000,    # 300s
    max_retries = 3,      # third times the charm
  )

#nbx /nbox_setup
